;
(define shift-key-mask
  (lambda (state)
    (= (bit-and state 1) 1)))
(define control-key-mask
  (lambda (state)
    (= (bit-and state 2) 2)))
(define alt-key-mask
  (lambda (state)
    (= (bit-and state 4) 4)))
(define meta-key-mask
  (lambda (state)
    (= (bit-and state 8) 8)))

(define modifier-key-mask
  (lambda (state)
    (> state 0)))
;;
(define backspace-key
  (lambda (key key-state)
    (= key 'backspace)))
(define left-key
  (lambda (key key-state)
    (= key 'left)))
(define right-key
  (lambda (key key-state)
    (= key 'right)))

(define modifier-key
  (lambda (key key-state)
    (or
     (= key 'Shift_key)
     (= key 'Alt_key)
     (= key 'Control_key)
     (= key 'Meta_key))))

(define switch-im-key
  (lambda (key state)
    (or
     (and (control-key-mask state)
	  (= key 'Shift_key))
     (and (shift-key-mask state)
	  (= key 'Control_key)))))

;;
(define parse-key-str
  (lambda (str translators key key-state)
    (let ((str-len (string-length str)))
      (cond
       ((= str-len 0)
	(list "" translators -1 key-state))
       ((= str-len 1)
	(list "" translators (string->charcode str) key-state))
       ((<= 2 str-len)
	(let* ((split (reverse (string-to-list str)))
	       (rest (if (= str-len 2)
			 ""
			 (string-list-concat (reverse (cddr split)))))
	       (modifier (string-append
			  (car split)
			  (cadr split))))
	  (cond
	   ((string-equal? modifier "S-") ; Shift
	    (set! key-state
		  (bit-or key-state 1)))
	   ((string-equal? modifier "C-") ; Control
	    (set! key-state
		  (bit-or key-state 2)))
	   ((string-equal? modifier "A-") ; Alt
	    (set! key-state
		  (bit-or key-state 4)))
	   ((string-equal? modifier "M-") ; Meta
	    (set! key-state
		  (bit-or key-state 8)))
	   ((string-equal? modifier "I-") ; case insensitive
	    (set! translators
		  (cons (lambda (key key-state)
			  (let ((translated-key (to-lower-char key)))
			    (list translated-key key-state)))
			translators)))
	   (else
	    (set! key (intern str))  ; keysym
	    (set! rest "")))
	  (if (string-equal? rest "")
	      (list "" translators key key-state)
	      (parse-key-str rest translators key key-state))))))))

(define translate-key-str
  (lambda (str translators key key-state)
    (if (not (string-equal? str ""))
	(let* ((parsed (parse-key-str str translators key key-state))
	       (parsed-keys (cddr parsed)))
	  (set! str "")
	  (set! translators (cadr parsed))
	  (set! key (car parsed-keys))
	  (set! key-state (cadr parsed-keys))))
    (if (null? translators)
	(list key key-state)
	(let* ((translator (car translators))
	       (rest-translators (cdr translators))
	       (translated (translator key key-state))
	       (translated-key (car translated))
	       (translated-state (cadr translated)))
	  (translate-key-str
	   str
	   rest-translators
	   translated-key
	   translated-state)))))

(define make-single-key-predicate
  (lambda (key-str)
    (let* ((translated (translate-key-str key-str () -1 0))
	   (target-key (car translated))
	   (target-state (cadr translated)))
      (lambda (key key-state)
	(and
	 (= key target-key)
	 (= key-state target-state))))))

(define make-key-predicate
  (lambda (key-strs)
    (let* ((key-strs (if (not (list? key-strs))
			 (list key-strs)
			 key-strs))
	   (predicates (mapcar make-single-key-predicate key-strs)))
      (lambda (key key-state)
	(apply proc-or
	       (mapcar (lambda (predicate)
			 (apply predicate (list key key-state)))
		       predicates))))))

(define set-key-binding
  (lambda (im state flavor)
    nil
    ))

(define regist-key-binding 
  (lambda (bind-name bind-table)
    nil
))
